@@ -4,9 +4,11 @@
 Basic objects for defining statistical inference models.
 """
 
+from __future__ import annotations
+
 import enum
 import copy as _copy
-from typing import Optional, Tuple, List, Union, Dict, Generator, Callable, TextIO, Sequence, Any
+from typing import Generator, Callable, TextIO, Sequence, Any
 
 import law
 import order as od
@@ -84,7 +86,7 @@ class ParameterTransformations(tuple):
     """
 
     def __new__(cls, transformations):
-        # TODO: at this point one could interfere and complain in case incompatible transfos are used
+        # TODO: at this point one could object / complain in case incompatible transfos are used
         transformations = [
             (t if isinstance(t, ParameterTransformation) else ParameterTransformation[t])
             for t in transformations
@@ -114,38 +116,38 @@ class InferenceModel(Derivable):
 
         categories:
           - name: cat1
-            category: 1e
-            variable: ht
-            data_datasets: [data_mu_a]
+            config_category: 1e
+            config_variable: ht
+            config_data_datasets: [data_mu_a]
             data_from_processes: []
             mc_stats: 10
             processes:
               - name: HH
-                process: hh
-                signal: True
-                mc_datasets: [hh_ggf]
+                config_process: hh
+                is_signal: True
+                config_mc_datasets: [hh_ggf]
                 parameters:
                   - name: lumi
                     type: rate_gauss
                     effect: 1.02
-                    shift_source: null
+                    config_shift_source: null
                   - name: pu
                     type: rate_gauss
                     effect: [0.97, 1.02]
-                    shift_source: null
+                    config_shift_source: null
                   - name: pileup
                     type: shape
                     effect: 1.0
-                    shift_source: minbias_xs
+                    config_shift_source: minbias_xs
               - name: tt
-                signal: False
-                process: ttbar
-                mc_datasets: [tt_sl, tt_dl, tt_fh]
+                is_signal: False
+                config_process: ttbar
+                config_mc_datasets: [tt_sl, tt_dl, tt_fh]
                 parameters:
                   - name: lumi
                     type: rate_gauss
                     effect: 1.02
-                    shift_source: null
+                    config_shift_source: null
 
           - name: cat2
             ...
@@ -218,31 +220,30 @@ class InferenceModel(Derivable):
     def category_spec(
         cls,
         name: str,
-        category: Optional[str] = None,
-        variable: Optional[str] = None,
-        data_datasets: Optional[Sequence[str]] = None,
-        data_from_processes: Optional[Sequence[str]] = None,
-        mc_stats: Optional[Union[float, tuple]] = None,
+        config_category: str | None = None,
+        config_variable: str | None = None,
+        config_data_datasets: Sequence[str] | None = None,
+        data_from_processes: Sequence[str] | None = None,
+        mc_stats: float | tuple | None = None,
     ) -> DotDict:
         """
         Returns a dictionary representing a category (interchangeably called bin or channel in other
         tools), forwarding all arguments.
 
             - *name*: The name of the category in the model.
-            - *category*: The name of the source category in the config to use. Note the possibly
-              ambiguous yet consistent naming.
-            - *variable*: The name of the variable in the config to use.
-            - *data_datasets*: List of names of datasets in the config to use for real data.
+            - *config_category*: The name of the source category in the config to use.
+            - *config_variable*: The name of the variable in the config to use.
+            - *config_data_datasets*: List of names of datasets in the config to use for real data.
             - *data_from_processes*: Optional list of names of :py:meth:`process_spec` objects that,
-              when *data_datasets* is not defined, make of a fake data contribution.
+              when *config_data_datasets* is not defined, make of a fake data contribution.
             - *mc_stats*: Either *None* to disable MC stat uncertainties, or a float or tuple of
               floats to control the options of MC stat options.
         """
         return DotDict([
             ("name", str(name)),
-            ("category", str(category) if category else None),
-            ("variable", str(variable) if variable else None),
-            ("data_datasets", list(map(str, data_datasets or []))),
+            ("config_category", str(config_category) if config_category else None),
+            ("config_variable", str(config_variable) if config_variable else None),
+            ("config_data_datasets", list(map(str, config_data_datasets or []))),
             ("data_from_processes", list(map(str, data_from_processes or []))),
             ("mc_stats", mc_stats),
             ("processes", []),
@@ -252,24 +253,23 @@ class InferenceModel(Derivable):
     def process_spec(
         cls,
         name: str,
-        process: Optional[str] = None,
-        signal: bool = False,
-        mc_datasets: Optional[Sequence[str]] = None,
+        config_process: str | None = None,
+        is_signal: bool = False,
+        config_mc_datasets: Sequence[str] | None = None,
     ) -> DotDict:
         """
         Returns a dictionary representing a process, forwarding all arguments.
 
             - *name*: The name of the process in the model.
-            - *process*: The name of the source process in the config to use. Note the possibly
-              ambiguous yet consistent naming.
-            - *signal*: A boolean flag deciding whether this process describes signal.
-            - *mc_datasets*: List of names of MC datasets in the config to use.
+            - *is_signal*: A boolean flag deciding whether this process describes signal.
+            - *config_process*: The name of the source process in the config to use.
+            - *config_mc_datasets*: List of names of MC datasets in the config to use.
         """
         return DotDict([
             ("name", str(name)),
-            ("process", str(process) if process else None),
-            ("signal", bool(signal)),
-            ("mc_datasets", list(map(str, mc_datasets or []))),
+            ("is_signal", bool(is_signal)),
+            ("config_process", str(config_process) if config_process else None),
+            ("config_mc_datasets", list(map(str, config_mc_datasets or []))),
             ("parameters", []),
         ])
 
@@ -277,10 +277,10 @@ class InferenceModel(Derivable):
     def parameter_spec(
         cls,
         name: str,
-        type: Union[ParameterType, str],
-        transformations: Sequence[Union[ParameterTransformation, str]] = (ParameterTransformation.none,),
-        shift_source: Optional[str] = None,
-        effect: Union[Any] = 1.0,
+        type: ParameterType | str,
+        transformations: Sequence[ParameterTransformation | str] = (ParameterTransformation.none,),
+        config_shift_source: str | None = None,
+        effect: Any | None = 1.0,
     ) -> DotDict:
         """
         Returns a dictionary representing a (nuisance) parameter, forwarding all arguments.
@@ -289,7 +289,7 @@ class InferenceModel(Derivable):
             - *type*: A :py:class:`ParameterType` instance describing the type of this parameter.
             - *transformations*: A sequence of :py:class:`ParameterTransformation` instances
               describing transformations to be applied to the effect of this parameter.
-            - *shift_source*: The name of a systematic shift source in the config that this
+            - *config_shift_source*: The name of a systematic shift source in the config that this
               parameter corresponds to.
             - *effect*: An arbitrary object describing the effect of the parameter (e.g. float for
               symmetric rate effects, 2-tuple for down/up variation, etc).
@@ -298,7 +298,7 @@ class InferenceModel(Derivable):
             ("name", str(name)),
             ("type", type if isinstance(type, ParameterType) else ParameterType[type]),
             ("transformations", ParameterTransformations(transformations)),
-            ("shift_source", str(shift_source) if shift_source else None),
+            ("config_shift_source", str(config_shift_source) if config_shift_source else None),
             ("effect", effect),
         ])
 
@@ -306,7 +306,7 @@ class InferenceModel(Derivable):
     def parameter_group_spec(
         cls,
         name: str,
-        parameter_names: Optional[Sequence[str]] = None,
+        parameter_names: Sequence[str] | None = None,
     ) -> DotDict:
         """
         Returns a dictionary representing a group of parameter names.
@@ -341,7 +341,7 @@ class InferenceModel(Derivable):
 
         # other cases are not supported
         raise Exception(
-            f"shape requirement cannot be evaluated of parameter '{param_obj.name}' with type "
+            f"shape requirement cannot be evaluated of parameter '{param_obj.name}' with type " +
             f"'{param_obj.type}' and transformations {param_obj.transformations}",
         )
 
@@ -358,7 +358,7 @@ class InferenceModel(Derivable):
         if callable(self.init_func):
             self.init_func()
 
-    def to_yaml(self, stream: Optional[TextIO] = None) -> Union[None, str]:
+    def to_yaml(self, stream: TextIO | None = None) -> str | None:
         """
         Writes the content of the :py:attr:`model` into a file-like object *stream* when given, and
         returns a string representation otherwise.
@@ -389,9 +389,9 @@ class InferenceModel(Derivable):
 
     def get_categories(
         self,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        category: str | Sequence[str] | None = None,
         only_names: bool = False,
-    ) -> List[Union[DotDict, str]]:
+    ) -> list[DotDict | str]:
         """
         Returns a list of categories whose name match *category*. *category* can be a string, a
         pattern, or sequence of them. When *only_names* is *True*, only names of categories are
@@ -409,10 +409,10 @@ class InferenceModel(Derivable):
 
     def get_category(
         self,
-        category: Union[str, Sequence[str]],
+        category: str | Sequence[str],
         only_name: bool = False,
         silent: bool = False,
-    ) -> Union[DotDict, str]:
+    ) -> DotDict | str:
         """
         Returns a single category whose name matches *category*. *category* can be a string, a
         pattern, or sequence of them. An exception is raised if no or more than one category is
@@ -438,7 +438,7 @@ class InferenceModel(Derivable):
 
     def has_category(
         self,
-        category: Union[str, Sequence[str]],
+        category: str | Sequence[str],
     ) -> bool:
         """
         Returns *True* if a category whose name matches *category* is existing, and *False*
@@ -468,7 +468,7 @@ class InferenceModel(Derivable):
 
     def remove_category(
         self,
-        category: Union[str, Sequence[str]],
+        category: str | Sequence[str],
     ) -> bool:
         """
         Removes one or more categories whose names match *category*. Returns *True* if at least one
@@ -499,11 +499,11 @@ class InferenceModel(Derivable):
 
     def get_processes(
         self,
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
         only_names: bool = False,
         flat: bool = False,
-    ) -> Union[Dict[str, Union[DotDict, str]], List[str]]:
+    ) -> dict[str, DotDict | str] | list[str]:
         """
         Returns a dictionary of processes whose names match *process*, mapped to the name of the
         category they belong to. Categories can optionally be filtered through *category*. Both
@@ -545,11 +545,11 @@ class InferenceModel(Derivable):
 
     def get_process(
         self,
-        process: Union[str, Sequence[str]],
-        category: Optional[Union[str, Sequence[str]]] = None,
+        process: str | Sequence[str],
+        category: str | Sequence[str] | None = None,
         only_name: bool = False,
         silent: bool = False,
-    ) -> Union[DotDict, str]:
+    ) -> DotDict | str:
         """
         Returns a single process whose name matches *process*, and optionally, whose category's name
         matches *category*. Both *process* and *category* can be a string, a pattern, or sequence of
@@ -590,7 +590,7 @@ class InferenceModel(Derivable):
                 return None
             names = processes if only_name else [p.name for p in processes]
             raise ValueError(
-                f"process '{process_name}' found more than once in category '{category}': "
+                f"process '{process_name}' found more than once in category '{category}': " +
                 ",".join(names),
             )
 
@@ -598,8 +598,8 @@ class InferenceModel(Derivable):
 
     def has_process(
         self,
-        process: Union[str, Sequence[str]],
-        category: Optional[Union[str, Sequence[str]]] = None,
+        process: str | Sequence[str],
+        category: str | Sequence[str] | None = None,
     ) -> bool:
         """
         Returns *True* if a process whose name matches *process*, and optionally whose category's
@@ -616,7 +616,7 @@ class InferenceModel(Derivable):
     def add_process(
         self,
         *args,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        category: str | Sequence[str] | None = None,
         silent: bool = False,
         **kwargs,
     ) -> None:
@@ -643,7 +643,7 @@ class InferenceModel(Derivable):
                 if silent:
                     continue
                 raise ValueError(
-                    f"process named '{process.name}' already registered in category "
+                    f"process named '{process.name}' already registered in category " +
                     f"'{category.name}'",
                 )
             target_categories.append(category)
@@ -654,8 +654,8 @@ class InferenceModel(Derivable):
 
     def remove_process(
         self,
-        process: Union[str, Sequence[str]],
-        category: Optional[Union[str, Sequence[str]]] = None,
+        process: str | Sequence[str],
+        category: str | Sequence[str] | None = None,
     ) -> bool:
         """
         Removes one or more processes whose names match *process*, and optionally whose category's
@@ -692,12 +692,12 @@ class InferenceModel(Derivable):
 
     def get_parameters(
         self,
-        parameter: Optional[Union[str, Sequence[str]]] = None,
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str] | None = None,
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
         only_names: bool = False,
         flat: bool = False,
-    ) -> Union[Dict[str, Dict[str, Union[DotDict, str]]], List[str]]:
+    ) -> dict[str, dict[str, DotDict | str]] | list[str]:
         """
         Returns a dictionary of parameter whose names match *parameter*, mapped twice to the name of
         the category and the name of the process they belong to. Categories and processes can
@@ -748,12 +748,12 @@ class InferenceModel(Derivable):
 
     def get_parameter(
         self,
-        parameter: Union[str, Sequence[str]],
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str],
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
         only_name: bool = False,
         silent: bool = False,
-    ) -> Union[DotDict, str]:
+    ) -> DotDict | str:
         """
         Returns a single parameter whose name matches *parameter*, and optionally, whose category's
         and process' name matches *category* and *process*. All three, *parameter*, *process* and
@@ -789,7 +789,7 @@ class InferenceModel(Derivable):
             if silent:
                 return None
             raise ValueError(
-                f"parameter '{parameter_name}' found in more than one category: "
+                f"parameter '{parameter_name}' found in more than one category: " +
                 ",".join(parameters),
             )
 
@@ -798,7 +798,7 @@ class InferenceModel(Derivable):
             if silent:
                 return None
             raise ValueError(
-                f"parameter '{parameter_name}' found in more than one process in category "
+                f"parameter '{parameter_name}' found in more than one process in category " +
                 f"'{category}': " + ",".join(parameters),
             )
 
@@ -808,7 +808,7 @@ class InferenceModel(Derivable):
                 return None
             names = parameters if only_name else [p.name for p in parameters]
             raise ValueError(
-                f"parameter '{parameter_name}' found more than once in category '{category}' and "
+                f"parameter '{parameter_name}' found more than once in category '{category}' and " +
                 f"process '{process}': " + ",".join(names),
             )
 
@@ -816,9 +816,9 @@ class InferenceModel(Derivable):
 
     def has_parameter(
         self,
-        parameter: Union[str, Sequence[str]],
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str],
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
     ) -> bool:
         """
         Returns *True* if a parameter whose name matches *parameter*, and optionally whose
@@ -841,8 +841,8 @@ class InferenceModel(Derivable):
     def add_parameter(
         self,
         *args,
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
         **kwargs,
     ) -> DotDict:
         """
@@ -868,7 +868,7 @@ class InferenceModel(Derivable):
             for process in _processes:
                 if self.has_parameter(parameter.name, process=process.name, category=category_name):
                     raise ValueError(
-                        f"parameter named '{parameter.name}' already registered for process "
+                        f"parameter named '{parameter.name}' already registered for process " +
                         f"'{process.name}' in category '{category_name}'",
                     )
 
@@ -881,9 +881,9 @@ class InferenceModel(Derivable):
 
     def remove_parameter(
         self,
-        parameter: Union[str, Sequence[str]],
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str],
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
     ) -> bool:
         """
         Removes one or more parameters whose names match *parameter*, and optionally whose
@@ -923,9 +923,9 @@ class InferenceModel(Derivable):
 
     def get_parameter_groups(
         self,
-        group: Optional[Union[str, Sequence[str]]] = None,
+        group: str | Sequence[str] | None = None,
         only_names: bool = False,
-    ) -> List[Union[DotDict, str]]:
+    ) -> list[DotDict | str]:
         """
         Returns a list of parameter group whose name match *group*. *group* can be a string, a
         pattern, or sequence of them.
@@ -945,9 +945,9 @@ class InferenceModel(Derivable):
 
     def get_parameter_group(
         self,
-        group: Union[str, Sequence[str]],
+        group: str | Sequence[str],
         only_name: bool = False,
-    ) -> Union[DotDict, str]:
+    ) -> DotDict | str:
         """
         Returns a single parameter group whose name matches *group*. *group* can be a string, a
         pattern, or sequence of them.
@@ -971,7 +971,7 @@ class InferenceModel(Derivable):
 
     def has_parameter_group(
         self,
-        group: Union[str, Sequence[str]],
+        group: str | Sequence[str],
     ) -> bool:
         """
         Returns *True* if a parameter group whose name matches *group* is existing, and *False*
@@ -1000,7 +1000,7 @@ class InferenceModel(Derivable):
 
     def remove_parameter_group(
         self,
-        group: Union[str, Sequence[str]],
+        group: str | Sequence[str],
     ) -> bool:
         """
         Removes one or more parameter groups whose names match *group*. *group* can be a string, a
@@ -1023,8 +1023,8 @@ class InferenceModel(Derivable):
 
     def add_parameter_to_group(
         self,
-        parameter: Union[str, Sequence[str]],
-        group: Union[str, Sequence[str]],
+        parameter: str | Sequence[str],
+        group: str | Sequence[str],
     ) -> bool:
         """
         Adds a parameter named *parameter* to one or multiple parameter groups whose name match
@@ -1062,8 +1062,8 @@ class InferenceModel(Derivable):
 
     def remove_parameter_from_groups(
         self,
-        parameter: Union[str, Sequence[str]],
-        group: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str],
+        group: str | Sequence[str] | None = None,
     ) -> bool:
         """
         Removes all parameters matching *parameter* from parameter groups whose names match *group*.
@@ -1098,8 +1098,8 @@ class InferenceModel(Derivable):
 
     def get_categories_with_process(
         self,
-        process: Union[str, Sequence[str]],
-    ) -> List[str]:
+        process: str | Sequence[str],
+    ) -> list[str]:
         """
         Returns a flat list of category names that contain processes matching *process*. *process*
         can be a string, a pattern, or sequence of them.
@@ -1112,10 +1112,10 @@ class InferenceModel(Derivable):
 
     def get_processes_with_parameter(
         self,
-        parameter: Union[str, Sequence[str]],
-        category: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str],
+        category: str | Sequence[str] | None = None,
         flat: bool = True,
-    ) -> Union[List[str], Dict[str, List[str]]]:
+    ) -> list[str] | dict[str, list[str]]:
         """
         Returns a dictionary of names of processes that contain a parameter whose names match
         *parameter*, mapped to categories names. Categories can optionally be filtered through
@@ -1147,10 +1147,10 @@ class InferenceModel(Derivable):
 
     def get_categories_with_parameter(
         self,
-        parameter: Union[str, Sequence[str]],
-        process: Optional[Union[str, Sequence[str]]] = None,
+        parameter: str | Sequence[str],
+        process: str | Sequence[str] | None = None,
         flat: bool = True,
-    ) -> Union[List[str], Dict[str, List[str]]]:
+    ) -> list[str] | dict[str, list[str]]:
         """
         Returns a dictionary of category names mapping to process names that contain parameters
         whose name match *parameter*. Processes can optionally be filtered through *process*. Both
@@ -1182,8 +1182,8 @@ class InferenceModel(Derivable):
 
     def get_groups_with_parameter(
         self,
-        parameter: Union[str, Sequence[str]],
-    ) -> List[str]:
+        parameter: str | Sequence[str],
+    ) -> list[str]:
         """
         Returns a list of names of parameter groups that contain a parameter whose name matches
         *parameter*, which can be a string, a pattern, or sequence of them.
@@ -1254,9 +1254,9 @@ class InferenceModel(Derivable):
 
     def iter_processes(
         self,
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
-    ) -> Generator[Tuple[DotDict, DotDict], None, None]:
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
+    ) -> Generator[tuple[DotDict, DotDict], None, None]:
         """
         Generator that iteratively yields all processes whose names match *process*, optionally
         in all categories whose names match *category*. The yielded value is a 2-tuple containing
@@ -1269,10 +1269,10 @@ class InferenceModel(Derivable):
 
     def iter_parameters(
         self,
-        parameter: Optional[Union[str, Sequence[str]]] = None,
-        process: Optional[Union[str, Sequence[str]]] = None,
-        category: Optional[Union[str, Sequence[str]]] = None,
-    ) -> Generator[Tuple[DotDict, DotDict, DotDict], None, None]:
+        parameter: str | Sequence[str] | None = None,
+        process: str | Sequence[str] | None = None,
+        category: str | Sequence[str] | None = None,
+    ) -> Generator[tuple[DotDict, DotDict, DotDict], None, None]:
         """
         Generator that iteratively yields all parameters whose names match *parameter*, optionally
         in all processes and categories whose names match *process* and *category*. The yielded
@@ -1286,10 +1286,10 @@ class InferenceModel(Derivable):
 
 
 def inference_model(
-    func: Optional[Callable] = None,
+    func: Callable | None = None,
     bases=(),
     **kwargs,
-) -> Union[DerivableMeta, Callable]:
+) -> DerivableMeta | Callable:
     """
     Decorator for creating a new :py:class:`InferenceModel` subclass with additional, optional
     *bases* and attaching the decorated function to it as ``init_func``. All additional *kwargs* are
