@@ -4,24 +4,46 @@
 Generic tools and base tasks that are defined along typical objects in an analysis.
 """
 
+from __future__ import annotations
+
 import os
 import enum
 import importlib
 import itertools
 import inspect
 import functools
-from typing import Optional, Sequence, List, Union, Set, Dict
+from typing import Sequence
 
 import luigi
 import law
 import order as od
 
+from columnflow.util import DotDict
+
 
 default_analysis = law.config.get_expanded("analysis", "default_analysis")
 default_config = law.config.get_expanded("analysis", "default_config")
 default_dataset = law.config.get_expanded("analysis", "default_dataset")
 
 
+class Requirements(DotDict):
+
+    def __init__(self, *others, **kwargs):
+        super().__init__()
+
+        # add others and kwargs
+        for reqs in others + (kwargs,):
+            self.update(reqs)
+
+
+class BaseTask(law.Task):
+
+    task_namespace = law.config.get_expanded("analysis", "cf_task_namespace", "cf")
+
+    # container for upstream requirements for convenience
+    reqs = Requirements()
+
+
 class OutputLocation(enum.Enum):
     """
     Output location flag.
@@ -32,11 +54,6 @@ class OutputLocation(enum.Enum):
     wlcg = "wlcg"
 
 
-class BaseTask(law.Task):
-
-    task_namespace = os.getenv("CF_TASK_NAMESPACE")
-
-
 class AnalysisTask(BaseTask, law.SandboxTask):
 
     analysis = luigi.Parameter(
@@ -48,24 +65,29 @@ class AnalysisTask(BaseTask, law.SandboxTask):
     allow_empty_sandbox = True
     sandbox = None
 
+    message_cache_size = 25
     local_workflow_require_branches = False
     output_collection_cls = law.SiblingFileCollection
 
     # defaults for targets
     default_store = "$CF_STORE_LOCAL"
     default_wlcg_fs = "wlcg_fs"
-    default_output_location = "local"
+    default_output_location = "config"
 
     @classmethod
-    def modify_param_values(cls, params):
-        params = super().modify_param_values(params)
-
+    def resolve_param_values(cls, params: dict) -> dict:
         # store a reference to the analysis inst
-        if "analysis" in params:
+        if "analysis_inst" not in params and "analysis" in params:
             params["analysis_inst"] = cls.get_analysis_inst(params["analysis"])
 
         return params
 
+    @classmethod
+    def modify_param_values(cls, params: dict) -> dict:
+        params = super().modify_param_values(params)
+        params = cls.resolve_param_values(params)
+        return params
+
     @classmethod
     def get_analysis_inst(cls, analysis):
         # prepare names
@@ -94,7 +116,11 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         """
         # always prefer certain parameters given as task family parameters (--TaskFamily-parameter)
         _prefer_cli = set(law.util.make_list(kwargs.get("_prefer_cli", [])))
-        _prefer_cli.add("version")
+        _prefer_cli |= {
+            "version", "workflow", "job_workers", "poll_interval", "walltime", "max_runtime",
+            "retries", "acceptance", "tolerance", "parallel_jobs", "shuffle_jobs", "htcondor_cpus",
+            "htcondor_gpus", "htcondor_pool",
+        }
         kwargs["_prefer_cli"] = _prefer_cli
 
         # when cls accepts a version, but non was actively requested, use the version map to assign it
@@ -132,9 +158,17 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         return ()
 
     @classmethod
-    def determine_allowed_shifts(cls, config_inst, params):
-        # implemented only for simplified mro control
-        return set()
+    def get_known_shifts(cls, config_inst: od.Config, params: dict) -> tuple[set[str], set[str]]:
+        """
+        Returns two sets of shifts in a tuple: shifts implemented by _this_ task, and depdenent
+        shifts implemented by upstream tasks.
+        """
+        # get shifts from upstream dependencies, consider both their own and upstream shifts as one
+        upstream_shifts = set()
+        for req in cls.reqs.values():
+            upstream_shifts |= set.union(*(req.get_known_shifts(config_inst, params) or (set(),)))
+
+        return set(), upstream_shifts
 
     @classmethod
     def get_array_function_kwargs(cls, task=None, **params):
@@ -144,30 +178,30 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         }
 
     @classmethod
-    def get_calibrator_kwargs(cls, task=None, **params):
+    def get_calibrator_kwargs(cls, *args, **kwargs):
         # implemented here only for simplified mro control
-        return cls.get_array_function_kwargs(task=task, **params)
+        return cls.get_array_function_kwargs(*args, **kwargs)
 
     @classmethod
-    def get_selector_kwargs(cls, task=None, **params):
+    def get_selector_kwargs(cls, *args, **kwargs):
         # implemented here only for simplified mro control
-        return cls.get_array_function_kwargs(task=task, **params)
+        return cls.get_array_function_kwargs(*args, **kwargs)
 
     @classmethod
-    def get_producer_kwargs(cls, task=None, **params):
+    def get_producer_kwargs(cls, *args, **kwargs):
         # implemented here only for simplified mro control
-        return cls.get_array_function_kwargs(task=task, **params)
+        return cls.get_array_function_kwargs(*args, **kwargs)
 
     @classmethod
     def find_config_objects(
         cls,
-        names: Union[str, Sequence[str], Set[str]],
+        names: str | Sequence[str] | set[str],
         container: od.UniqueObject,
         object_cls: od.UniqueObjectMeta,
-        object_groups: Optional[Dict[str, list]] = None,
+        object_groups: dict[str, list] | None = None,
         accept_patterns: bool = True,
         deep: bool = False,
-    ) -> List[str]:
+    ) -> list[str]:
         """
         Returns all names of objects of type *object_cls* known to a *container* (e.g.
         :py:class:`od.Analysis` or :py:class:`od.Config`) that match *names*. A name can also be a
@@ -209,7 +243,7 @@ class AnalysisTask(BaseTask, law.SandboxTask):
 
         object_names = []
         patterns = []
-        lookup = list(names)
+        lookup = law.util.make_list(names)
         while lookup:
             name = lookup.pop(0)
             if has_obj(name):
@@ -263,29 +297,40 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         return parts
 
     def local_path(self, *path, **kwargs):
-        """ local_path(*path, store=None)
-        Joins path fragments from *store* (defaulting to :py:attr:`default_store`), *store_parts()*
-        and *path* and returns the joined path.
+        """ local_path(*path, store=None, fs=None)
+        Joins path fragments from *store* (defaulting to :py:attr:`default_store`),
+        :py:meth:`store_parts` and *path* and returns the joined path. In case a *fs* is defined,
+        it should refer to the config section of a local file system, and consequently, *store* is
+        not prepended to the returned path as the resolution of absolute paths is handled by that
+        file system.
         """
-        # determine the main store directory
-        store = kwargs.get("store") or self.default_store
+        # if no fs is set, determine the main store directory
+        parts = ()
+        if not kwargs.pop("fs", None):
+            store = kwargs.get("store") or self.default_store
+            parts += (store,)
 
         # concatenate all parts that make up the path and join them
-        parts = tuple(self.store_parts().values()) + path
-        path = os.path.join(store, *(str(p) for p in parts))
+        parts += tuple(self.store_parts().values()) + path
+        path = os.path.join(*map(str, parts))
 
         return path
 
     def local_target(self, *path, **kwargs):
-        """ local_target(*path, dir=False, store=None, **kwargs)
+        """ local_target(*path, dir=False, store=None, fs=None, **kwargs)
         Creates either a local file or directory target, depending on *dir*, forwarding all *path*
-        fragments and *store* to :py:meth:`local_path` and all *kwargs* the respective target class.
+        fragments, *store* and *fs* to :py:meth:`local_path` and all *kwargs* the respective target
+        class.
         """
+        _dir = kwargs.pop("dir", False)
+        store = kwargs.pop("store", None)
+        fs = kwargs.get("fs", None)
+
         # select the target class
-        cls = law.LocalDirectoryTarget if kwargs.pop("dir", False) else law.LocalFileTarget
+        cls = law.LocalDirectoryTarget if _dir else law.LocalFileTarget
 
         # create the local path
-        path = self.local_path(*path, store=kwargs.pop("store", None))
+        path = self.local_path(*path, store=store, fs=fs)
 
         # create the target instance and return it
         return cls(path, **kwargs)
@@ -299,7 +344,7 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         """
         # concatenate all parts that make up the path and join them
         parts = tuple(self.store_parts().values()) + path
-        path = os.path.join(*(str(p) for p in parts))
+        path = os.path.join(*map(str, parts))
 
         return path
 
@@ -309,11 +354,12 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         *path* fragments to :py:meth:`wlcg_path` and all *kwargs* the respective target class. When
         *None*, *fs* defaults to the *default_wlcg_fs* class level attribute.
         """
+        _dir = kwargs.pop("dir", False)
         if not kwargs.get("fs"):
             kwargs["fs"] = self.default_wlcg_fs
 
         # select the target class
-        cls = law.wlcg.WLCGDirectoryTarget if kwargs.pop("dir", False) else law.wlcg.WLCGFileTarget
+        cls = law.wlcg.WLCGDirectoryTarget if _dir else law.wlcg.WLCGFileTarget
 
         # create the local path
         path = self.wlcg_path(*path)
@@ -333,8 +379,8 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         if location == OutputLocation.config:
             location = law.config.get_expanded("outputs", self.task_family, split_csv=True)
             if not location:
-                self.logger.info(
-                    f"no option 'outputs.{self.task_family}' found in law.cfg to obtain target "
+                self.logger.debug(
+                    f"no option 'outputs::{self.task_family}' found in law.cfg to obtain target "
                     "location, falling back to 'local'",
                 )
                 location = ["local"]
@@ -344,11 +390,12 @@ class AnalysisTask(BaseTask, law.SandboxTask):
         # forward to correct function
         if location[0] == OutputLocation.local:
             # get other options
-            (store,) = (location[1:] + [None])[:1]
-            kwargs.setdefault("store", store)
+            (loc,) = (location[1:] + [None])[:1]
+            loc_key = "fs" if (loc and law.config.has_section(loc)) else "store"
+            kwargs.setdefault(loc_key, loc)
             return self.local_target(*path, **kwargs)
 
-        elif location[0] == OutputLocation.wlcg:
+        if location[0] == OutputLocation.wlcg:
             # get other options
             (fs,) = (location[1:] + [None])[:1]
             kwargs.setdefault("fs", fs)
@@ -365,11 +412,11 @@ class ConfigTask(AnalysisTask):
     )
 
     @classmethod
-    def modify_param_values(cls, params):
-        params = super().modify_param_values(params)
+    def resolve_param_values(cls, params: dict) -> dict:
+        params = super().resolve_param_values(params)
 
         # store a reference to the config inst
-        if "analysis_inst" in params and "config" in params:
+        if "config_inst" not in params and "analysis_inst" in params and "config" in params:
             params["config_inst"] = params["analysis_inst"].get_config(params["config"])
 
         return params
@@ -412,15 +459,13 @@ class ShiftTask(ConfigTask):
         description="name of a systematic shift to apply; must fulfill order.Shift naming rules; "
         "default: 'nominal'",
     )
-    effective_shift = luigi.Parameter(default=law.NO_STR)
+    local_shift = luigi.Parameter(default=law.NO_STR)
 
-    # shifts implemented by this task or one of its requirements
-    shifts = set()
-
-    # skip passing effective_shift to cli completion, req params and sandboxing
-    exclude_params_index = {"effective_shift"}
-    exclude_params_req = {"effective_shift"}
-    exclude_params_sandbox = {"effective_shift"}
+    # skip passing local_shift to cli completion, req params and sandboxing
+    exclude_params_index = {"local_shift"}
+    exclude_params_req = {"local_shift"}
+    exclude_params_sandbox = {"local_shift"}
+    exclude_params_remote_workflow = {"local_shift"}
 
     allow_empty_shift = False
 
@@ -432,90 +477,100 @@ class ShiftTask(ConfigTask):
         return params
 
     @classmethod
-    def modify_param_values(cls, params):
-        """
-        When "config" and "shift" are set, this method evaluates them to set the effecitve shift.
-        For that, it takes the shifts stored in the config instance and compares it with those
-        defined by this class.
-        """
-        # the modify_param_values super method must not necessarily be set
-        super_func = super().modify_param_values
-        if callable(super_func):
-            params = super_func(params)
+    def resolve_param_values(cls, params: dict) -> dict:
+        params = super().resolve_param_values(params)
 
-        # get params
-        config_inst = params.get("config_inst")
-        requested_shift = params.get("shift")
-        requested_effective_shift = params.get("effective_shift")
-
-        # initialize the effective shift when not set
-        no_values = (None, law.NO_STR)
-        if requested_effective_shift in no_values:
-            params["effective_shift"] = "nominal"
-
-        # do nothing when the effective shift is already set and no config is defined
-        if requested_effective_shift not in no_values or config_inst in no_values:
-            return params
-
-        # shift must be set
-        if requested_shift in no_values:
-            if cls.allow_empty_shift:
-                return params
-            raise Exception(f"no shift found in params: {params}")
-
-        # complain when the requested shift is not known
-        if requested_shift not in config_inst.shifts:
-            raise ValueError(f"unknown shift: {requested_shift}")
-
-        # determine the allowed shifts for this class
-        allowed_shifts = cls.determine_allowed_shifts(config_inst, params)
-
-        # when allowed, add it to the task parameters
-        if requested_shift in allowed_shifts:
-            params["effective_shift"] = requested_shift
-
-        # store references
-        params["shift_inst"] = config_inst.get_shift(requested_shift)
-        params["effective_shift_inst"] = config_inst.get_shift(params["effective_shift"])
+        # set default shift
+        if params.get("shift") in (None, law.NO_STR):
+            params["shift"] = "nominal"
 
         return params
 
     @classmethod
-    def determine_allowed_shifts(cls, config_inst, params):
-        # for the basic shift task, only the shifts implemented by this task class are allowed
-        # still call super for simplified mro control
-        shifts = super().determine_allowed_shifts(config_inst, params)
+    def modify_param_values(cls, params):
+        """
+        When "config" and "shift" are set, this method evaluates them to set the global shift.
+        For that, it takes the shifts stored in the config instance and compares it with those
+        defined by this class.
+        """
+        params = super().modify_param_values(params)
 
-        # add class level shifts
-        shifts |= cls.shifts
+        # get params
+        config_inst = params.get("config_inst")
+        requested_shift = params.get("shift")
+        requested_local_shift = params.get("local_shift")
 
-        return shifts
+        # require that the config is set
+        if config_inst in (None, law.NO_STR):
+            return params
+
+        # require that the shift is set and known
+        if requested_shift in (None, law.NO_STR):
+            if cls.allow_empty_shift:
+                params["shift"] = law.NO_STR
+                params["local_shift"] = law.NO_STR
+                return params
+            raise Exception(f"no shift found in params: {params}")
+        if requested_shift not in config_inst.shifts:
+            raise ValueError(f"shift {requested_shift} unknown to {config_inst}")
+
+        # determine the known shifts for this class
+        shifts, upstream_shifts = cls.get_known_shifts(config_inst, params)
+
+        # actual shift resolution: compare the requested shift to known ones
+        # local_shift -> the requested shift if implemented by the task itself, else nominal
+        # shift       -> the requested shift if implemented by this task
+        #                or an upsteam task (== global shift), else nominal
+        if requested_local_shift in (None, law.NO_STR):
+            if requested_shift in shifts:
+                params["shift"] = requested_shift
+                params["local_shift"] = requested_shift
+            elif requested_shift in upstream_shifts:
+                params["shift"] = requested_shift
+                params["local_shift"] = "nominal"
+            else:
+                params["shift"] = "nominal"
+                params["local_shift"] = "nominal"
+
+        # store references
+        params["global_shift_inst"] = config_inst.get_shift(params["shift"])
+        params["local_shift_inst"] = config_inst.get_shift(params["local_shift"])
+
+        return params
 
     @classmethod
     def get_array_function_kwargs(cls, task=None, **params):
         kwargs = super().get_array_function_kwargs(task=task, **params)
+
         if task:
-            if task.shift_inst:
-                kwargs["shift_inst"] = task.shift_inst
-        elif "effective_shift" in params and "config_inst" in kwargs:
-            if params["effective_shift"] != law.NO_STR:
-                kwargs["shift_inst"] = kwargs["config_inst"].get_shift(params["effective_shift"])
+            if task.local_shift_inst:
+                kwargs["shift_inst"] = task.local_shift_inst
+            if task.global_shift_inst:
+                kwargs["global_shift_inst"] = task.global_shift_inst
+        else:
+            if "local_shift_inst" in params:
+                kwargs["local_shift_inst"] = params["local_shift_inst"]
+            if "global_shift_inst" in params:
+                kwargs["global_shift_inst"] = params["global_shift_inst"]
+
         return kwargs
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
-        # store a reference to the effective shift instance
-        self.shift_inst = None
-        if self.effective_shift and self.effective_shift != law.NO_STR:
-            self.shift_inst = self.config_inst.get_shift(self.effective_shift)
+        # store references to the shift instances
+        self.local_shift_inst = None
+        self.global_shift_inst = None
+        if self.shift not in (None, law.NO_STR) and self.local_shift not in (None, law.NO_STR):
+            self.global_shift_inst = self.config_inst.get_shift(self.shift)
+            self.local_shift_inst = self.config_inst.get_shift(self.local_shift)
 
     def store_parts(self):
         parts = super().store_parts()
 
         # add the shift name
-        if self.shift_inst:
-            parts.insert_after("config", "shift", self.shift_inst.name)
+        if self.global_shift_inst:
+            parts.insert_after("config", "shift", self.global_shift_inst.name)
 
         return parts
 
@@ -530,15 +585,32 @@ class DatasetTask(ShiftTask):
     file_merging = None
 
     @classmethod
-    def modify_param_values(cls, params):
-        params = super().modify_param_values(params)
+    def resolve_param_values(cls, params):
+        params = super().resolve_param_values(params)
 
         # store a reference to the dataset inst
-        if "config_inst" in params and "dataset" in params:
+        if "dataset_inst" not in params and "config_inst" in params and "dataset" in params:
             params["dataset_inst"] = params["config_inst"].get_dataset(params["dataset"])
 
         return params
 
+    @classmethod
+    def get_known_shifts(cls, config_inst: od.Config, params: dict) -> tuple[set[str], set[str]]:
+        # dataset can have shifts, that are considered as upstream shifts
+        shifts, upstream_shifts = super().get_known_shifts(config_inst, params)
+
+        dataset_inst = params.get("dataset_inst")
+        if dataset_inst:
+            if dataset_inst.is_data:
+                # clear all shifts for data
+                shifts.clear()
+                upstream_shifts.clear()
+            else:
+                # extend with dataset variations for mc
+                upstream_shifts |= set(dataset_inst.info.keys())
+
+        return shifts, upstream_shifts
+
     @classmethod
     def get_version_params(cls):
         params = super().get_version_params()
@@ -548,24 +620,6 @@ class DatasetTask(ShiftTask):
             params += ("dataset",)
         return params
 
-    @classmethod
-    def determine_allowed_shifts(cls, config_inst, params):
-        # dataset can have shifts, so extend the set of allowed shifts
-        allowed_shifts = super().determine_allowed_shifts(config_inst, params)
-
-        # dataset must be set
-        if "dataset" in params:
-            requested_dataset = params.get("dataset")
-            if requested_dataset not in (None, law.NO_STR):
-                dataset_inst = config_inst.get_dataset(requested_dataset)
-                # clear shifts for data and extend with dataset variations for mc
-                if dataset_inst.is_data:
-                    allowed_shifts.clear()
-                else:
-                    allowed_shifts |= set(dataset_inst.info.keys())
-
-        return allowed_shifts
-
     @classmethod
     def get_array_function_kwargs(cls, task=None, **params):
         kwargs = super().get_array_function_kwargs(task=task, **params)
@@ -581,8 +635,12 @@ class DatasetTask(ShiftTask):
         # store references to the dataset instance
         self.dataset_inst = self.config_inst.get_dataset(self.dataset)
 
-        # store dataset info for the effective shift
-        key = self.effective_shift if self.effective_shift in self.dataset_inst.info else "nominal"
+        # store dataset info for the global shift
+        key = (
+            self.global_shift_inst.name
+            if self.global_shift_inst and self.global_shift_inst.name in self.dataset_inst.info else
+            "nominal"
+        )
         self.dataset_info_inst = self.dataset_inst.get_info(key)
 
     def store_parts(self):
@@ -631,9 +689,10 @@ class DatasetTask(ShiftTask):
         """
         Hook to modify the additional info printed along logs of the htcondor workflow.
         """
+        info.append(self.config_inst.name)
         info.append(self.dataset_inst.name)
-        if self.shift_inst and self.shift_inst.name != "nominal":
-            info.append(self.shift_inst.name)
+        if self.global_shift_inst not in (None, law.NO_STR, "nominal"):
+            info.append(self.global_shift_inst.name)
         return info
 
 
@@ -765,8 +824,8 @@ def wrapper_factory(
     base_cls: law.Task,
     require_cls: AnalysisTask,
     enable: Sequence[str],
-    cls_name: Optional[str] = None,
-    attributes: Optional[dict] = None,
+    cls_name: str | None = None,
+    attributes: dict | None = None,
 ) -> law.task.base.Register:
     """
     Factory function creating wrapper task classes, inheriting from *base_cls* and
@@ -820,6 +879,10 @@ def wrapper_factory(
                 f"'{','.join(known_features)}'",
             )
 
+    # treat base_cls as a tuple
+    base_classes = law.util.make_tuple(base_cls)
+    base_cls = base_classes[0]
+
     # define wrapper feature flags
     has_configs = "configs" in enable
     has_skip_configs = has_configs and "skip_configs" in enable
@@ -855,7 +918,7 @@ def wrapper_factory(
         check_class_compatibility("datasets", DatasetTask, ShiftTask)
 
     # create the class
-    class Wrapper(base_cls, law.WrapperTask):
+    class Wrapper(*base_classes, law.WrapperTask):
 
         if has_configs:
             configs = law.CSVParameter(
@@ -863,6 +926,7 @@ def wrapper_factory(
                 description="names or name patterns of configs to use; can also be the key of a "
                 "mapping defined in the 'config_groups' auxiliary data of the analysis; default: "
                 "('*',)",
+                brace_expand=True,
             )
         if has_skip_configs:
             skip_configs = law.CSVParameter(
@@ -870,6 +934,7 @@ def wrapper_factory(
                 description="names or name patterns of configs to skip after evaluating --configs; "
                 "can also be the key of a mapping defined in the 'config_groups' auxiliary data "
                 "of the analysis; empty default",
+                brace_expand=True,
             )
         if has_datasets:
             datasets = law.CSVParameter(
@@ -877,6 +942,7 @@ def wrapper_factory(
                 description="names or name patterns of datasets to use; can also be the key of a "
                 "mapping defined in the 'dataset_groups' auxiliary data of the corresponding "
                 "config; default: ('*',)",
+                brace_expand=True,
             )
         if has_skip_datasets:
             skip_datasets = law.CSVParameter(
@@ -884,6 +950,7 @@ def wrapper_factory(
                 description="names or name patterns of datasets to skip after evaluating "
                 "--datasets; can also be the key of a mapping defined in the 'dataset_groups' "
                 "auxiliary data of the corresponding config; empty default",
+                brace_expand=True,
             )
         if has_shifts:
             shifts = law.CSVParameter(
@@ -891,6 +958,7 @@ def wrapper_factory(
                 description="names or name patterns of shifts to use; can also be the key of a "
                 "mapping defined in the 'shift_groups' auxiliary data of the corresponding "
                 "config; default: ('nominal',)",
+                brace_expand=True,
             )
         if has_skip_shifts:
             skip_shifts = law.CSVParameter(
@@ -898,6 +966,7 @@ def wrapper_factory(
                 description="names or name patterns of shifts to skip after evaluating --shifts; "
                 "can also be the key of a mapping defined in the 'shift_groups' auxiliary data "
                 "of the corresponding config; empty default",
+                brace_expand=True,
             )
 
         def __init__(self, *args, **kwargs):
@@ -1002,7 +1071,7 @@ def wrapper_factory(
                     )
                     if not datasets:
                         raise ValueError(
-                            f"no datasets found in config {self.config_inst} matching "
+                            f"no datasets found in config {config_inst} matching "
                             f"{self.datasets}",
                         )
                     if self.wrapper_has_skip_datasets:
@@ -1015,7 +1084,7 @@ def wrapper_factory(
                         datasets = [d for d in datasets if d not in skip_datasets]
                         if not datasets:
                             raise ValueError(
-                                f"no datasets found in config {self.config_inst} after skipping "
+                                f"no datasets found in config {config_inst} after skipping "
                                 f"{self.skip_datasets}",
                             )
                     prod_sequences.append(sorted(datasets))
