@@ -4,16 +4,18 @@
 Collection of general helpers and utilities.
 """
 
+from __future__ import annotations
+
 __all__ = [
     "UNSET", "env_is_remote", "env_is_dev", "primes",
-    "maybe_import", "import_plt", "import_ROOT", "create_random_name", "expand_path", "real_path",
-    "ensure_dir", "wget", "call_thread", "call_proc", "ensure_proxy", "dev_sandbox", "safe_div",
-    "test_float", "is_pattern", "is_regex", "pattern_matcher", "get_root_processes_from_campaign",
+    "maybe_import", "import_plt", "import_ROOT", "import_file", "create_random_name", "expand_path",
+    "real_path", "ensure_dir", "wget", "call_thread", "call_proc", "ensure_proxy", "dev_sandbox",
+    "safe_div", "test_float", "is_pattern", "is_regex", "pattern_matcher",
+    "dict_add_strict", "get_source_code",
     "DotDict", "MockModule", "FunctionArgs", "ClassPropertyDescriptor", "classproperty",
     "DerivableMeta", "Derivable",
 ]
 
-
 import os
 import uuid
 import queue
@@ -27,11 +29,12 @@ import multiprocessing
 import multiprocessing.pool
 from functools import wraps
 from collections import OrderedDict
-from typing import Tuple, Callable, Any, Optional, Union, Sequence
+from typing import Callable, Any, Sequence, Union
 from types import ModuleType
 
 import law
-import order as od
+import luigi
+
 
 #: Placeholder for an unset value.
 UNSET = object()
@@ -57,9 +60,9 @@ primes = [
 
 def maybe_import(
     name: str,
-    package: Optional[str] = None,
+    package: str | None = None,
     force: bool = False,
-) -> Union[ModuleType, "MockModule"]:
+) -> ModuleType | MockModule:
     """
     Calls *importlib.import_module* internally and returns the module if it exists, or otherwise a
     :py:class:`MockModule` instance with the same name. When *force* is *True* and the import fails,
@@ -67,8 +70,10 @@ def maybe_import(
     """
     try:
         return importlib.import_module(name, package)
-    except ImportError:
-        if force:
+    except ImportError as e:
+        # raise in case force is set, or an other package than the requested one was not found
+        m = re.match(r"^No\smodule\snamed\s\'(.+)\'$", str(e))
+        if force or not m or not name.startswith(m.group(1)):
             raise
         if package:
             name = package + name
@@ -118,6 +123,30 @@ def import_ROOT() -> ModuleType:
     return _ROOT
 
 
+def import_file(path: str, attr: str | None = None):
+    """
+    Loads the content of a python file located at *path* and returns its package content as a
+    dictionary. When *attr* is set, only the attribute with that name is returned.
+
+    The file is not required to be importable as its content is loaded directly into the
+    interpreter. While this approach is not necessarily clean, it can be useful in places where
+    custom code must be loaded.
+    """
+    # load the package contents (do not try this at home)
+    path = expand_path(path)
+    pkg = DotDict()
+    with open(path, "r") as f:
+        exec(f.read(), pkg)
+
+    # extract a particular attribute
+    if attr:
+        if attr not in pkg:
+            raise AttributeError(f"no local member '{attr}' found in file {path}")
+        return pkg[attr]
+
+    return pkg
+
+
 def create_random_name() -> str:
     """
     Returns a random string based on UUID v4.
@@ -192,9 +221,9 @@ def wget(src: str, dst: str, force: bool = False) -> str:
 def call_thread(
     func: Callable,
     args: tuple = (),
-    kwargs: Optional[dict] = None,
-    timeout: Optional[float] = None,
-) -> Tuple[bool, Any, Optional[str]]:
+    kwargs: dict | None = None,
+    timeout: float | None = None,
+) -> tuple[bool, Any, str | None]:
     """
     Execute a function *func* in a thread and aborts the call when *timeout* is reached. *args* and
     *kwargs* are forwarded to the function.
@@ -223,9 +252,9 @@ def call_thread(
 def call_proc(
     func: Callable,
     args: tuple = (),
-    kwargs: Optional[dict] = None,
-    timeout: Optional[float] = None,
-) -> Tuple[bool, Any, Optional[str]]:
+    kwargs: dict | None = None,
+    timeout: float | None = None,
+) -> tuple[bool, Any, str | None]:
     """
     Execute a function *func* in a process and aborts the call when *timeout* is reached. *args* and
     *kwargs* are forwarded to the function.
@@ -259,14 +288,18 @@ def ensure_proxy(
     task: law.Task,
     *args,
     **kwargs,
-) -> Tuple[Callable, Callable, Callable]:
+) -> tuple[Callable, Callable, Callable]:
     """
     Law task decorator that checks whether either a voms or arc proxy is existing before calling
     the decorated method.
     """
     def before_call():
-        # do nothing for grid jobs
-        if env_is_grid:
+        # do nothing in remote jobs
+        if env_is_remote:
+            return None
+
+        # do nothing when explicitly skipped by the law config
+        if law.config.get_expanded_boolean("analysis", "skip_ensure_proxy", default=False):
             return None
 
         # check the proxy validity
@@ -307,8 +340,11 @@ def dev_sandbox(sandbox: str) -> str:
     if _type not in ["venv", "bash"]:
         return sandbox
 
-    # create the dev path and check if it exists
-    dev_path = "{}_dev{}".format(*os.path.splitext(path))
+    # create the dev path and check if it exists, with special treatment of the cf_{prod,dev} envs
+    if path == "$CF_BASE/sandboxes/cf_prod.sh":
+        dev_path = "$CF_BASE/sandboxes/cf_dev.sh"
+    else:
+        dev_path = "{}_dev{}".format(*os.path.splitext(path))
     if not os.path.exists(real_path(dev_path)):
         return sandbox
 
@@ -350,7 +386,7 @@ def memoize(f: Callable) -> Callable:
     return wrapper
 
 
-def safe_div(a: Union[int, float], b: Union[int, float]) -> float:
+def safe_div(a: int | float, b: int | float) -> float:
     """
     Returns *a* divided by *b* if *b* is not zero, and zero otherwise.
     """
@@ -384,7 +420,7 @@ def is_regex(s: str) -> bool:
     return s.startswith("^") and s.endswith("$")
 
 
-def pattern_matcher(pattern: Union[Sequence[str], str], mode=any) -> Callable[[str], bool]:
+def pattern_matcher(pattern: Sequence[str] | str, mode: Callable = any) -> Callable[[str], bool]:
     r"""
     Takes a string *pattern* which might be an actual pattern for fnmatching, a regular expressions
     or just a plain string and returns a function that can be used to test of a string matches that
@@ -435,24 +471,35 @@ def pattern_matcher(pattern: Union[Sequence[str], str], mode=any) -> Callable[[s
     return lambda s: s == pattern
 
 
-def get_root_processes_from_campaign(campaign: od.Campaign) -> od.UniqueObjectIndex:
+def dict_add_strict(d: dict, key: str, value: Any) -> None:
     """
-    Extracts all root process objects from datasets contained in an order campaign and returns them
-    in a unique object index.
+    Adds key-value pair to dictionary, but only if it does not change an existing value;
+    Raises KeyError otherwise.
     """
-    # get all dataset processes
-    processes = set.union(*map(set, (dataset.processes for dataset in campaign.datasets)))
+    if key in d.keys() and d[key] != value:
+        raise KeyError(f"'{d.__class__.__name__}' object already has key {key}")
+    d[key] = value
 
-    # get their root processes
-    root_processes = set.union(*map(set, (process.get_root_processes() for process in processes)))
 
-    # create an empty index and fill subprocesses via walking
-    index = od.UniqueObjectIndex(cls=od.Process)
-    for root_process in root_processes:
-        for process, _, _ in root_process.walk_processes(include_self=True):
-            index.add(process, overwrite=True)
+def get_source_code(obj: Any, indent: str | int = None) -> str:
+    """
+    Returns the source code of any object *obj* as a string. When *indent* is not *None*, the code
+    indentation is first removed and then re-applied with *indent* if it is a string, or by that
+    many spaces in case it is an integer.
+    """
+    code = inspect.getsource(obj)
 
-    return index
+    if indent is not None:
+        code = code.replace("\t", "    ")
+        lines = code.split("\n")
+        n_old_indent = len(lines[0]) - len(lines[0].lstrip(" "))
+        new_indent = (" " * indent) if isinstance(indent, int) else indent
+        code = "\n".join(
+            (new_indent + line[n_old_indent:]) if line.strip() else ""
+            for line in lines
+        )
+
+    return code
 
 
 class DotDict(OrderedDict):
@@ -488,21 +535,21 @@ class DotDict(OrderedDict):
         # => 1
     """
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str) -> Any:
         try:
             return self[attr]
         except KeyError:
             raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")
 
-    def __setattr__(self, attr, value):
+    def __setattr__(self, attr: str, value: Any) -> None:
         self[attr] = value
 
-    def copy(self):
+    def copy(self) -> DotDict:
         """"""
         return self.__class__(self)
 
     @classmethod
-    def wrap(cls, *args, **kwargs) -> "DotDict":
+    def wrap(cls, *args, **kwargs) -> DotDict:
         """
         Takes a dictionary *d* and recursively replaces it and all other nested dictionary types
         with :py:class:`DotDict`'s for deep attribute-style access.
@@ -533,26 +580,30 @@ class MockModule(object):
        The name of the mock module.
     """
 
-    def __init__(self, name):
+    def __init__(self, name: str):
         super().__init__()
 
         self._name = name
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str) -> MockModule:
         return self
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"<{self.__class__.__name__} '{self._name}' at {hex(id(self))}>"
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args, **kwargs) -> None:
         raise Exception(f"{self._name} is a mock module and cannot be called")
 
-    def __nonzero__(self):
+    def __nonzero__(self) -> bool:
         return False
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
         return False
 
+    def __or__(self, other) -> Any:
+        # forward union type hints
+        return Union[type(self), other]
+
 
 class FunctionArgs(object):
     """
@@ -576,13 +627,13 @@ class ClassPropertyDescriptor(object):
     Generic descriptor class that is used by :py:func:`classproperty`.
     """
 
-    def __init__(self, fget: Callable, fset: Optional[Callable] = None):
+    def __init__(self, fget: Callable, fset: Callable | None = None):
         super().__init__()
 
         self.fget = fget
         self.fset = fset
 
-    def __get__(self, obj: type, cls: Optional[type] = None) -> Any:
+    def __get__(self, obj: type, cls: type | None = None) -> Any:
         if cls is None:
             cls = type(obj)
 
@@ -596,7 +647,7 @@ class ClassPropertyDescriptor(object):
         return self.fset.__get__(obj, type(obj))(value)
 
 
-def classproperty(func):
+def classproperty(func: Callable) -> ClassPropertyDescriptor:
     """
     Propety decorator for class-level methods.
     """
@@ -620,7 +671,7 @@ class DerivableMeta(type):
         cls_name: str,
         bases: tuple,
         cls_dict: dict,
-    ) -> "DerivableMeta":
+    ) -> DerivableMeta:
         """
         Class creation.
         """
@@ -657,7 +708,7 @@ class DerivableMeta(type):
         cls_name: str,
         deep: bool = True,
         silent: bool = False,
-    ) -> Union["DerivableMeta", None]:
+    ) -> DerivableMeta | None:
         """
         Returns a previously created subclass named *cls_name*. When *deep* is *True*, the lookup is
         recursive through all levels of subclasses. When no such subclass was found an exception is
@@ -690,8 +741,9 @@ class DerivableMeta(type):
         cls,
         cls_name: str,
         bases: tuple = (),
-        cls_dict: Optional[dict] = None,
-    ) -> "DerivableMeta":
+        cls_dict: dict | None = None,
+        module: str | None = None,
+    ) -> DerivableMeta:
         """
         Creates a subclass named *cls_name* inheriting from *this* class an additional, optional
         *bases*. *cls_dict* will be attached as class-level attributes.
@@ -703,14 +755,15 @@ class DerivableMeta(type):
         subcls = cls.__class__(cls_name, (cls,) + bases, cls_dict or {})
 
         # overwrite __module__ to point to the module of the calling stack
-        frame = inspect.stack()[1]
-        module = inspect.getmodule(frame[0])
+        if not module:
+            frame = inspect.stack()[1]
+            module = inspect.getmodule(frame[0])
         if module:
             subcls.__module__ = module.__name__
 
         return subcls
 
-    def derived_by(cls, other: "DerivableMeta") -> bool:
+    def derived_by(cls, other: DerivableMeta) -> bool:
         """
         Returns if a class *other* is either *this* or derived from *this* class, and *False*
         otherwise.
@@ -722,14 +775,64 @@ class Derivable(object, metaclass=DerivableMeta):
     """
     Derivable base class with features provided by the meta :py:class:`DerivableMeta`.
 
-    .. py:attribute:: cls_name
+    .. py:classattribute:: cls_name
        type: str
        read-only
 
        A shorthand to access the name of the class.
     """
 
-    @property
-    def cls_name(self) -> str:
+    @classproperty
+    def cls_name(cls) -> str:
         # shorthand to the class name
-        return self.__class__.__name__
+        return cls.__name__
+
+
+class KeyValueMessage(luigi.worker.SchedulerMessage):
+    """
+    Subclass of :py:class:`luigi.worker.SchedulerMessage` that adds :py:attr:`key` and
+    :py:attr:`value` attributes, parsed from the incoming message assuming a format ``key = value``.
+
+    .. py:attribute: key
+       type: str
+
+       The key of the message.
+
+    .. py:attribute: value
+       type: str
+
+       The value of the message.
+    """
+
+    # compile expression for key - value parsing of scheduler messages
+    message_cre = re.compile(r"^\s*([^\=\:]+)\s*(\=|\:)\s*(.*)\s*$")
+
+    @classmethod
+    def from_message(cls, message: luigi.worker.SchedulerMessage) -> KeyValueMessage | None:
+        """
+        Factory for :py:class:`KeyValueMessage` instances that takes an existing *message* object
+        and splits its content into a key value pair. The instance is returned if the parsing is
+        successful, and *None* otherwise.
+        """
+        m = cls.message_cre.match(message.content)
+        if not m:
+            return None
+
+        return cls(
+            message._scheduler,
+            message._task_id,
+            message._message_id,
+            message.content,
+            m.group(1),
+            m.group(3),
+            **message.payload,
+        )
+
+    def __init__(self, *args, key, value, **kwargs):
+        super().__init__(*args, **kwargs)
+
+        self.key = key
+        self.value = value
+
+    def __str__(self) -> str:
+        return str(self.value)
